#!/usr/bin/env python3
#
# Analyzer for umaudemc's test output
#

import argparse
import math
import itertools
import sys

#
## Constants of the problem
#

# All entries in the table that correspond to the input data
problem_data = ['filename', 'module', 'initial', 'formula', 'strategy', 'opaque', 'expected', 'logic']
# All backends that may appear in the input
backends = ['maude', 'ltsmin', 'pymc', 'nusmv', 'spot', 'builtin']

# All logics supported by each backend
supported_logics = {
	'LTL':    {'maude', 'ltsmin', 'pymc', 'nusmv', 'spot'},
	'CTL':    {'ltsmin', 'pymc', 'nusmv', 'builtin'},
	'CTL*':   {'ltsmin', 'pymc'},
	'Mucalc': {'ltsmin', 'builtin'}
}

# Renaming of logic names for pretty printing
logic_print = {
	'Mucalc': 'μ-calculus',
	'propLogic': 'propositional'
}

# Renaming of backend names for pretty printing
backend_print = {
	'maude': 'Maude',
	'ltsmin': 'LTSmin',
	'pymc': 'pyMC',
	'nusmv': 'NuSMV',
	'spot': 'Spot',
	'builtin': 'builtin'
}

# Backends that use the automata-theoretic approach
buchi_backends = ('maude', 'spot', 'ltsmin')


#
## Classes describing measures
#

class MeasureType:
	"""Type of a measure with potential parameters"""

	# Name of a logic
	LOGIC_T = 0
	# Name of a backend
	BACKEND_T = 1

	# Single real value
	VALUE_T = 0
	# Percentage
	PERC_T = 1
	# Dictionary
	DICT_T = 2

	# Greater is better
	BT_GREATER = 0
	# Lower is better
	BT_LOWER = 1
	# This measure is not comparable
	BT_NOCOMP = 2

	def __init__(self, name, sign, vtype, desc, better):
		# Name of the measure
		self.name = name
		# Signature of the measure
		self.sign = sign
		# Type of measure value
		self.vtype = vtype
		# Human-readable description of the measure
		self.desc = desc
		# Which results are better
		self.better = better

	def describe(self, args):
		"""Helper to produce an human-readable of measures of this type"""
	
		args_visible = [None] * len(args)

		for i in range(len(args)):
			if self.sign[i] == self.LOGIC_T:
				args_visible[i] = logic_print.get(args[i], args[i])
			else:
				args_visible[i] = backend_print.get(args[i], args[i])

		return self.desc.format(*args_visible)


class Measure:
	"""Measure of the model-checking executions"""

	def __init__(self, mtype, value, args=()):
		# Measure type
		self.mtype = mtype
		# Arguments of the measure (for parametric ones)
		self.args = args
		# Value of the measure
		self.value = value

	def __repr__(self):
		if self.args:
			args_str = '(' + ', '.join(self.args) + ')'
		else:
			args_str = ''
		return f'{self.mtype.name}{args_str} = {self.value}'

	def describe(self):
		"""Human-readable description of the measure"""
	
		return self.mtype.describe(self.args)

	@property
	def perc(self):
		"""Percentage measure"""

		if self.mtype.vtype == MeasureType.VALUE_T:
			return self.value

		return self.value[0] / self.value[1]

	def better(self, other):
		"""Whether the current measure is better than other"""
	
		greater = None

		if self.mtype.vtype == MeasureType.VALUE_T:
			greater = self.value - other.value

		elif self.mtype.vtype == MeasureType.PERC_T:
			greater = self.perc - other.perc

		if greater is None or self.mtype.better == MeasureType.BT_NOCOMP:
			return None

		else:
			return greater if self.mtype.better == MeasureType.BT_GREATER else - greater


class MeasureFamily:
	"""Family of parametric measures"""

	def __init__(self, summary, name, sign, mtype, desc, better=MeasureType.BT_GREATER):
		self.summary = summary
		self.type = MeasureType(name, sign, mtype, desc, better)

	def value(self, args, *value):
		if len(value) == 1:
			value = value[0]
		else:
			value = tuple(value)

		self.summary.add_measure(args=args, value=value, mtype=self.type)


class Summary:
	"""Summary of statistical measures"""

	def __init__(self):
		self.sections = {}
		self.sections_desc = {}
		self.current = None

	def section(self, name, desc):
		self.current = self.sections.setdefault(name, {})
		self.sections_desc.setdefault(name, desc)
		return self.current

	def add_measure(self, name=None, args=(), vtype=None, value=None, desc=None, mtype=None, better=False):
		if mtype is None:
			mtype = MeasureType(name, (), vtype, desc, better)
		else:
			name = mtype.name

		mlist = self.current.setdefault(name, [])
		mlist.append(Measure(mtype, value, args))

	def stat(self, name, desc, value, better=MeasureType.BT_GREATER):
		self.add_measure(name=name, desc=desc, value=value, vtype=MeasureType.VALUE_T, better=better)

	def perc(self, name, desc, value, total, better=MeasureType.BT_GREATER):
		self.add_measure(name=name, desc=desc, value=(value, total), vtype=MeasureType.PERC_T, better=better)

	def dict(self, name, desc, value, better=MeasureType.BT_GREATER):
		self.add_measure(name=name, desc=desc, value=value, vtype=MeasureType.DICT_T, better=better)

	def family(self, name, desc, sign, vtype, better=MeasureType.BT_GREATER):
		return MeasureFamily(self, name, sign, vtype, desc, better)


def read_benchmark(input_file, aggout=None):
	# Load the execution log from the CSV file
	executions = pd.read_csv(input_file)

	# Calculate the mean of the multiple execution for the same case
	executions_means = executions.drop('result', axis=1).groupby(problem_data + ['backend'], as_index=True, dropna=False).mean()
	# Pass the backend from the rows to the columns and flatten the column combinations
	cases = executions_means.unstack()
	cases.columns = (f'{backend}_{prop.replace("_ms", "")}' for prop, backend in cases.columns.values)
	# Make the index column be data columns
	cases.reset_index(inplace=True)

	# Imprime los resultados a un archivo
	if aggout is not None:
		cases.to_csv(aggout, index=False)

	return executions, cases


def analyze_output(executions, cases):
	# A summary is build with some measures
	sm = Summary()

	sm.section('global', 'Global information')

	sm.stat('number-executions', 'Number of executions',
	        len(executions.index))

	failed_executions = executions.expected != executions.result
	failed_executions_count = failed_executions.sum()

	sm.perc('failed-executions', 'Failed executions',
	        failed_executions_count, len(executions.index))

	fam = sm.family('failed-backends', 'Wrong by backend {}', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T)

	for logic, value in executions[failed_executions].backend.value_counts().iteritems():
		fam.value((logic, ), value)

	#if failed_executions_count > 0 :
	#	print(executions[failed_executions][problem_data + ['backend', 'result']].drop(['module', 'initial'], axis=1).drop_duplicates())

	sm.stat('execution-time', 'Total execution time (min)',
	        executions.time_ms.sum() / (1000 * 60),
	        better=MeasureType.BT_LOWER)

	total_cases = len(cases.index)
	strategy_free_cases = cases.strategy.isnull().sum()

	sm.stat('number-cases', 'Number of cases',
	        total_cases)
	sm.perc('strategy-cases', 'Strategy-controlled cases',
	        total_cases - strategy_free_cases, total_cases,
	        better=MeasureType.BT_NOCOMP)
	sm.perc('classical-cases', 'Strategy-free cases',
	        strategy_free_cases, total_cases,
	        better=MeasureType.BT_NOCOMP)
	sm.perc('true-properties', 'True properties',
	        cases.expected.sum(), total_cases,
	        better=MeasureType.BT_NOCOMP)

	fam = sm.family('cases-logic', 'Cases for {}', (MeasureType.LOGIC_T, ), MeasureType.VALUE_T)

	for logic, value in cases.value_counts(subset='logic').iteritems():
		fam.value((logic, ), value)

	# Cases by logic
	cases_logic = { logic: cases[cases.logic == logic] for logic in supported_logics.keys() }
	# Number of cases by logic
	num_cases_logic = { logic: len(table.index) for logic, table in cases_logic.items() }

	sm.section('missing-cases', 'Missing cases per backend')

	fam = sm.family('missing-cases', 'Missing {} for {}',
	                (MeasureType.LOGIC_T, MeasureType.BACKEND_T), MeasureType.PERC_T,
	                better=MeasureType.BT_LOWER)

	for logic, sbackends in supported_logics.items():
		for backend in sbackends:
			fam.value((logic, backend),
			          cases_logic[logic][f'{backend}_time'].isnull().sum(), num_cases_logic[logic])

	# States and real states among backends
	sm.section('real-states', 'Real states')

	strat_cases = cases[cases.strategy.isnull() == False]
	strat_ltl_cases = strat_cases[strat_cases.logic == 'LTL']
	num_strat_cases = len(strat_cases.index)
	num_strat_ltl_cases = len(strat_ltl_cases.index)

	fam_cases = sm.family('has-unreal', 'Real states lower ({})', (MeasureType.BACKEND_T, ), MeasureType.PERC_T, better=MeasureType.BT_LOWER)
	fam_wperc = sm.family('weighted-unreal', 'Weighted percertage ({})', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_mean_inc = sm.family('mean-unreal', 'Mean increment ({})', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_max_inc = sm.family('max-unreal', 'Max increment ({})', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)

	# Real states variation per backend
	for backend in backends:
		backend_domain = strat_cases[f'{backend}_time'].isnull() == False

		less_real = strat_cases[backend_domain & (strat_cases[f'{backend}_states'] != strat_cases[f'{backend}_real_states'])]
		increment = less_real[f'{backend}_states'] - less_real[f'{backend}_real_states']
		num_backend_cases = backend_domain.sum()

		# Is the same as weighted
		global_perc = (less_real[f'{backend}_real_states']).sum() / (less_real[f'{backend}_states']).sum()

		fam_cases.value((backend, ), len(less_real.index), num_backend_cases)
		fam_wperc.value((backend, ), global_perc * 100)
		fam_mean_inc.value((backend, ), increment.mean())
		fam_max_inc.value((backend, ), increment.max())

	ltsmin_maude_real = strat_ltl_cases[(strat_ltl_cases.maude_states != strat_ltl_cases.maude_real_states) & (strat_ltl_cases.ltsmin_states == strat_ltl_cases.ltsmin_real_states)]
	maude_ltsmin_real = strat_ltl_cases[(strat_ltl_cases.maude_states == strat_ltl_cases.maude_real_states) & (strat_ltl_cases.ltsmin_states != strat_ltl_cases.ltsmin_real_states)]

	sm.perc('exact-ltsmin-not-maude', 'Exact LTSmin not Maude',
	        len(ltsmin_maude_real.index), num_strat_ltl_cases)
	sm.stat('mean-increment-maude',
	        'Mean increment', (ltsmin_maude_real.maude_states - ltsmin_maude_real.maude_real_states).mean())
	sm.perc('exact-maude-not-ltsmin', 'Exact Maude not LTSmin',
	        len(maude_ltsmin_real.index), num_strat_ltl_cases)
	sm.stat('mean-increment-ltsmin', 'Mean increment',
	        (maude_ltsmin_real.ltsmin_states - maude_ltsmin_real.ltsmin_real_states).mean())
#	print(maude_ltsmin_real[['filename', 'expected', 'ltsmin_time', 'maude_time', 'maude_states', 'maude_real_states', 'ltsmin_states', 'ltsmin_real_states']])
	sm.perc('states-maude>ltsmin', 'States Maude > LTSmin',
	        (strat_ltl_cases.maude_states > strat_ltl_cases.ltsmin_states).sum(), num_strat_ltl_cases)
	sm.perc('states-ltsmin>maude', 'States LTSmin > Maude',
	        (strat_ltl_cases.maude_states < strat_ltl_cases.ltsmin_states).sum(), num_strat_ltl_cases)#
	sm.stat('mean-maude-ltsmin', 'Mean Maude - LTSmin',
	        (strat_ltl_cases.maude_states - strat_ltl_cases.ltsmin_states).mean())

	sltlt_cases = strat_ltl_cases[strat_ltl_cases.expected]
	num_sltlt_cases = len(sltlt_cases.index)

#	sm.perc('States Maude > LTSmin (true)', (sltlt_cases.maude_states > sltlt_cases.ltsmin_states).sum(), num_sltlt_cases)
#	sm.perc('States LTSmin > Maude (true)', (sltlt_cases.maude_states < sltlt_cases.ltsmin_states).sum(), num_sltlt_cases)
#	sm.stat('Mean Maude - LTSmin (true)', (sltlt_cases.maude_states - sltlt_cases.ltsmin_states).mean())


	# Preparation time versus total time
	sm.section('preparation-time', 'Preparation time per backend')

	fam_mean = sm.family('preparation-perc-mean', 'Preparation {} (mean, %)', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_weig = sm.family('preparation-perc-weig', 'Preparation {} (weighted, %)', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_max = sm.family('preparation-perc-max', 'Preparation {} (max, %)', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_min = sm.family('preparation-perc-min', 'Preparation {} (min, %)', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)

	for backend in backends:
		prep_cases = cases[cases[f'{backend}_prep'].isnull() == False]

		if prep_cases.empty:
			continue

		time_quo = (prep_cases[f'{backend}_prep'] / prep_cases[f'{backend}_time']) * 100
		weighted_mean = prep_cases[f'{backend}_prep'].sum() / prep_cases[f'{backend}_time'].sum()

		fam_mean.value((backend, ), time_quo.mean())
		fam_weig.value((backend, ), weighted_mean * 100)
		fam_max.value((backend, ), time_quo.max())
		fam_min.value((backend, ), time_quo.min())

	# Length of the counterexamples
	sm.section('counterexamples', 'Counterexamples')

	fam_mean    = sm.family('counterlength-mean', 'Counter length {}', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_common  = sm.family('counterlength-common', 'Counter length {} (common)', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T, better=MeasureType.BT_LOWER)
	fam_shorter = sm.family('shorter-counter', '{} < {}', (MeasureType.BACKEND_T, MeasureType.BACKEND_T), MeasureType.PERC_T, better=MeasureType.BT_LOWER)

	# In principle, all backends must return an LTL counterexample
	nusmv_counter = cases_logic['LTL'][cases_logic['LTL'].nusmv_clength.isnull() == False]

	for backend in backends:
		counter_cases = cases[cases[f'{backend}_clength'].isnull() == False]

		if counter_cases.empty:
			continue

		fam_mean.value((backend, ), nusmv_counter[f'{backend}_clength'].mean())
		fam_common.value((backend, ), counter_cases[f'{backend}_clength'].mean())

		if backend != 'maude':
			fam_shorter.value((backend, 'maude'), (counter_cases.maude_clength > counter_cases[f'{backend}_clength']).sum(), len(counter_cases.index))
			fam_shorter.value(('maude', backend), (nusmv_counter.maude_clength < nusmv_counter[f'{backend}_clength']).sum(), len(nusmv_counter.index))


	# Büchi automaton
	sm.section('buchi', 'Büchi automata')

	buchi_data = executions[executions.buchi.isnull() == False][problem_data + ['backend', 'buchi']]
	buchi_data = buchi_data.groupby(problem_data + ['backend'], as_index=False, dropna=False).mean()

	fam_mean = sm.family('buchi-mean', 'Mean Büchi states ({})', (MeasureType.BACKEND_T, ), MeasureType.VALUE_T)

	for backend in buchi_backends:
		fam_mean.value((backend, ), buchi_data[buchi_data.backend == backend].buchi.mean())

	#for lhs, rhs in itertools.permutations(buchi_backends, 2):
	#	printer.stat(f'{lhs} < {rhs}', (buchi_data[buchi_data.backend == lhs].buchi < buchi_data[buchi_data.backend == rhs].buchi).sum())

	return sm


def make_top(index, values):
	"""Calculate the most significant differences"""

	x = [(value, index[0]) for index, value in zip(index, values)]
	x.sort(reverse=True)
	return x


def compare_executions(cases1, cases2):
	"""Compare changes in cases from different executions"""

	joined = cases1.set_index(problem_data).join(cases2.set_index(problem_data), how='inner', lsuffix='1', rsuffix='2')

	print('\n*** More rewrites (in Maude) ***')

	tmp = joined[joined.maude_rewrites2 > joined.maude_rewrites1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_rewrites2 - tmp.maude_rewrites1)))

	print('\n*** Less rewrites (in Maude) ***')

	tmp = joined[joined.maude_rewrites2 < joined.maude_rewrites1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_rewrites1 - tmp.maude_rewrites2)))

	print('\n*** More states (in Maude) ***')

	tmp = joined[joined.maude_states2 > joined.maude_states1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_states2 - tmp.maude_states1)))

	print('\n*** Less states (in Maude) ***')

	tmp = joined[joined.maude_states2 < joined.maude_states1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_states1 - tmp.maude_states2)))

	print('\n*** More time (in Maude) ***')

	tmp = joined[joined.maude_time2 > joined.maude_time1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_time2 - tmp.maude_time1)))

	print('\n*** Less time (in Maude) ***')

	tmp = joined[joined.maude_time2 < joined.maude_time1]
	print(len(tmp), make_top(list(tmp.index), list(tmp.maude_time1 - tmp.maude_time2)))


def print_summary(summary):
	"""Print a summary"""

	for name, section in summary.sections.items():
		print(f'\n***** {summary.sections_desc[name]} *****\n')
		for cases in section.values():
			for case in cases:
				if case.mtype.vtype == MeasureType.VALUE_T:
					print(f'{case.describe():40}: {case.value}')

				elif case.mtype.vtype == MeasureType.PERC_T:
					print(f'{case.describe():40}: {case.value[0]:<30} {round(case.perc * 100, 2)} %')


def print_comparison(summary1, summary2):
	"""Print the comparison between two summaries"""

	for sname, section in summary1.sections.items():
		print(f'\n***** {summary1.sections_desc[sname]} *****\n')

		section2 = summary2.sections[sname]

		for mname, cases in section.items():
			cases2 = section2[mname]

			for case1, case2 in itertools.product(cases, cases2):
				if case1.args != case2.args:
					continue

				better = case2.better(case1)

				if better is not None and better != 0:
					sys.stdout.write('\x1b[33m' if better > 0 else '\x1b[31m')

					if case1.mtype.vtype == MeasureType.VALUE_T:
						print(f'{case1.describe():40}: {case1.value} {case2.value}')

					elif case1.mtype.vtype == MeasureType.PERC_T:
						print(f'{case1.describe():40}: {round(case1.perc * 100, 2):30} {round(case2.perc * 100, 2):30}')

					sys.stdout.write('\x1b[0m')


def graph_output(executions, cases):
	"""Draw graph with the results"""

	from matplotlib.backends.backend_pdf import PdfPages

	cases.sort_values(by='ltsmin_states', inplace=True)
	plt_def = dict(linestyle='none', fillstyle='none')

	with PdfPages('output.pdf') as pdf:
		# Logics pie chart
		logic_share = cases.logic.value_counts()
		logic_share.plot.pie(ylabel='Logic', labels=(logic_print.get(l, l) for l in logic_share.index))
		pdf.savefig()
		plt.close()
		
		# Expected output pie chart
		cases.expected.value_counts().plot.pie(ylabel='Result')
		pdf.savefig()
		plt.close()
		
		# Verification time (linear)
		ltl_cases = cases[((cases.logic == 'LTL') | (cases.logic == 'propLogic')) & cases.strategy.notna() & (cases.ltsmin_states > 2)]
		initc = lambda s: (s - s.min()) / (ltl_cases.ltsmin_time - ltl_cases.ltsmin_time.min())
		
		plt.plot(ltl_cases.ltsmin_states, initc(ltl_cases.maude_time), color='black')
		plt.plot(ltl_cases.ltsmin_states, initc(ltl_cases.nusmv_time), marker='^', color='b', **plt_def)
		plt.plot(ltl_cases.ltsmin_states, initc(ltl_cases.pymc_time), marker='s', color='r', **plt_def)
		plt.plot(ltl_cases.ltsmin_states, initc(ltl_cases.spot_time), marker='+', color='lime', **plt_def)
		ax = plt.gca()
		ax.set_xscale('log')
		plt.ylim([-.5, 14.5])
		plt.xlabel('Number of states', labelpad=8.0)
		plt.ylabel('Time proportion respect to LTSmin', labelpad=8.0)
		plt.hlines(1, 0, 2e5, color='yellow', linestyles='dotted')
		plt.title('LTL', pad=10.0)
		plt.legend(('Maude', 'NuSMV', 'pyModelChecking', 'Spot'))
		#plt.savefig('ltl.pgf')
		pdf.savefig()
		plt.close()

		# Verification time (branching)
		btime_cases = cases[(cases.logic != 'LTL') & (cases.logic != 'propLogic') & (cases.ltsmin_states > 5)]
		initc = lambda s: (s - s.min()) / (btime_cases.ltsmin_time - btime_cases.ltsmin_time.min())
		
		plt.plot(btime_cases.ltsmin_states, initc(btime_cases.nusmv_time), marker='^', color='b', **plt_def)
		plt.plot(btime_cases.ltsmin_states, initc(btime_cases.pymc_time), marker='s', color='r', **plt_def)
		plt.plot(btime_cases.ltsmin_states, initc(btime_cases.builtin_time), marker='+', color='lime', **plt_def)
		ax = plt.gca()
		ax.set_xscale('log')
		plt.ylim([-.1, 2.3])
		plt.xlabel('Number of states', labelpad=8.0)
		plt.ylabel('Time proportion respect to LTSmin', labelpad=10.0)
		plt.legend(('NuSMV', 'pyModelChecking', 'Builtin'))
		#plt.savefig('branching.pgf')
		pdf.savefig()
		plt.close()


if __name__ == '__main__':
	parser = argparse.ArgumentParser(description='Utility to analyze the test results produced by umaduemc')
	parser.add_argument('input', help='Test execution log as produced by umaudemc (CSV file).')
	parser.add_argument('second', help='Second test execution log to compare (CSV file).', nargs='?')
	parser.add_argument('--graph', help='Generate graphs instead of a text summary', action='store_true')
	parser.add_argument('-o', help='Write results aggregated by test case to a CSV file.')

	args = parser.parse_args()

	# Since NumPy, Pandas and Matplotlib take some time to load,
	# we defer it until the arguments have been checked
	import pandas as pd
	import numpy as np
	import matplotlib.pyplot as plt

	# Read the benchmarks from the file
	executions, cases = read_benchmark(args.input, aggout=args.o)

	if args.second is not None:
		executions2, cases2 = read_benchmark(args.second)
		summary1 = analyze_output(executions, cases)
		summary2 = analyze_output(executions2, cases2)
		print_comparison(summary1, summary2)
		compare_executions(cases, cases2)

	elif args.graph:
		graph_output(executions, cases)

	else:
		summary = analyze_output(executions, cases)
		print_summary(summary)
